//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//						(c) 1997 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	image.cc -- implementation of the Image class
//
//	CHANGE HISTORY:
//              DAH, 25/03/98: 1.0.0: - first version
//              DAH, 06/04/98: 1.0.1: - make _copyGlobalAttribs method; bug fix for
//                                      image extraction.
//				UL, 2000-12-20: 2.0.0: - couple of trivial changes to make the
//										 optimizer happy
//
//	AUTHOR:
//              D. Hinshaw
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Expanded description of SAS task `evselect'
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/task_devel/packages/evselect/src/image.cc,v 1.35 2000/12/22 13:36:12 ulammers Exp $

#include <iostream>
#include <strstream>
#include <cmath>
#include <values.h>

#include "param.h"
#include "Dal.h"
#include "errstr.h"
#include <ExprColumnGenerator.h>
#include "Extractor.h"
#include "DataSubspace.h"
#include "wcs.h"

void Image::make(bool writeDss){

  bool zcol = booleanParameter("withzcolumn");

  // Get pointers to the needed columns

  Column *xDatCol, *yDatCol, *zDatCol = 0;
  Column *xCol, *yCol, *zCol = 0;

  std::string xcolumn = stringParameter("xcolumn");
  if( _intable->hasColumn(xcolumn) ){
    xCol = _intable->column(xcolumn);
    xDatCol=reType(xCol, Dal::Real64);
  }else{
    std::strstream newName;
    newName << "X" << _intable << ends;
    ExprColumnGenerator colGenX(xcolumn, newName.str());
    colGenX.doIt(*_intable, *_tempTable);
    xCol = xDatCol = _intable->column(xcolumn);
  }

  std::string ycolumn = stringParameter("ycolumn");
  if( _intable->hasColumn(ycolumn) ){
    yCol = _intable->column(ycolumn);
    yDatCol=reType(yCol, Dal::Real64);
  }else{
    std::strstream newName;
    newName << "Y" << _intable << ends;
    ExprColumnGenerator colGenY(ycolumn, newName.str());
    colGenY.doIt(*_intable, *_tempTable);
    yCol = yDatCol = _intable->column(ycolumn);
  }

  if(zcol){
    std::string zcolumn = stringParameter("zcolumn");
    if( _intable->hasColumn(zcolumn) ){
      zCol = _intable->column(zcolumn);
      zDatCol=reType(zCol, Dal::Real64);
    }else{
      std::strstream newName;
      newName << "Z" << _intable << ends;
      ExprColumnGenerator colGenZ(zcolumn, newName.str());
      colGenZ.doIt(*_intable, *_tempTable);
      zCol = zDatCol = _intable->column(zcolumn);
    }
  }

  bool xIntCol = xCol->dataType()==Dal::Int8 || xCol->dataType()==Dal::Int16 || 
    xCol->dataType()==Dal::Int32;
  bool yIntCol = yCol->dataType()==Dal::Int8 || yCol->dataType()==Dal::Int16 || 
    yCol->dataType()==Dal::Int32;
  bool zIntCol = false;
  if(zcol) zIntCol = zCol->dataType()==Dal::Int8 || zCol->dataType()==Dal::Int16 || 
    zCol->dataType()==Dal::Int32;

  // Determine the size and binning factor of the image

  std::vector<real64> xLimits(2), yLimits(2);

  GetLimits(xCol, "ximagemin", "ximagemax", booleanParameter("withxranges"), xLimits);
  GetLimits(yCol, "yimagemin", "yimagemax", booleanParameter("withyranges"), yLimits);

  // For integer valued columns, adjust for suitably different meaning fo limits

  if( xIntCol ){
    if( xLimits[0] != floor(xLimits[0]) || xLimits[1] != floor(xLimits[1]) ){
      errstr << "X ranges must be integers for integer-valued columns."<<Sas::error("XRangeVal");
    }
    xLimits[0] -= 0.5;
    xLimits[1] += 0.5;
  }

  if( yIntCol ){
    if( yLimits[0] != floor(yLimits[0]) || yLimits[1] != floor(yLimits[1]) ){
      errstr << "Y ranges must be integers for integer-valued columns."<<Sas::error("YRangeVal");
    }
    yLimits[0] -= 0.5;
    yLimits[1] += 0.5;
  }

  if(booleanParameter("withcelestialcenter")){
    vector<double> world;
    world.push_back(realParameter("raimagecenter"));
    world.push_back(realParameter("decimagecenter"));

    vector<double> pixelCenter = _radec2xy(world, xCol, yCol);
    double xHalf = (xLimits[1]-xLimits[0])/2;
    double yHalf = (yLimits[1]-yLimits[0])/2;

    xLimits[0] = pixelCenter[0]-xHalf;
    xLimits[1] = pixelCenter[0]+xHalf;
    yLimits[0] = pixelCenter[1]-yHalf;
    yLimits[1] = pixelCenter[1]+yHalf;
  }


  if( xLimits[1] <= xLimits[0] || yLimits[1] <= yLimits[0] ){

    if( xLimits[1] == xLimits[0] || yLimits[1] == yLimits[0] )
      errstr << "One of the columns for extracting the image has "
             << "the same value throughout, or the upper and lower limits "
	     << "for this column are equal. No image will "
	     << "be created." << Sas::warning("BadImgLimits"); 
    else 
      errstr << "Upper image limit is less than lower limit.  No image will "
	     << "be created." << Sas::warning("BadImgLimits");

    return;
  }

  std::vector<real64> binSize(2);
  std::vector<long unsigned> size;

  if( stringParameter("imagebinning")=="binSize" ){

    binSize[0] = realParameter("yimagebinsize");
    binSize[1] = realParameter("ximagebinsize");
    if( binSize[0]==0 || binSize[1]==0 )
      errstr << "Image bin size set to zero.  This does not make sense." <<Sas::error("ImgZeroBinVal");


    if( yIntCol && floor(binSize[0]) != binSize[0] ||
	xIntCol && floor(binSize[1]) != binSize[1] ){
      errstr << "Binning factor must be an integer for integer-valued columns."<<Sas::error("ImgBinVal");
    }

    size.push_back( ceil(float(yLimits[1]-yLimits[0])/binSize[0]) );
    size.push_back( ceil(float(xLimits[1]-xLimits[0])/binSize[1]) );

  }else{

    size.push_back( intParameter("yimagesize") );
    size.push_back( intParameter("ximagesize") );
    if( size[0]==0 || size[1]==0 )
      errstr << "Image size set to zero.  This does not make sense." <<Sas::error("ImgZeroSizeVal");

    if(yIntCol){
      binSize[0] = ceil( (yLimits[1]-yLimits[0])/float(size[0]) );
    }else{
      errstr << "Y axis upper limit adjusted by an epsilon amount." << Sas::warning("YlimitEpsilon");
      yLimits[1] *= 1.0 + 1.0E-10;
      binSize[0] = (yLimits[1]-yLimits[0])/float(size[0]);
    }

    if(xIntCol){
      binSize[1] = ceil( (xLimits[1]-xLimits[0])/float(size[1]) );
    }else{
      errstr << "X axis upper limit adjusted by an epsilon amount." << Sas::warning("XlimitEpsilon");
      xLimits[1] *= 1.0 + 1.0E-10;
      binSize[1] = (xLimits[1]-xLimits[0])/float(size[1]);
    }

    if( booleanParameter("squarepixels") ){
      if( binSize[1] > binSize[0] ) binSize[0] = binSize[1];
      else binSize[1] = binSize[0];
    }

    setParameter("ximagebinsize", binSize[1]);
    setParameter("yimagebinsize", binSize[0]);

    if(yIntCol) size[0] = (long unsigned)ceil((yLimits[1]-yLimits[0])/binSize[0]);
    if(xIntCol) size[1] = (long unsigned)ceil((xLimits[1]-xLimits[0])/binSize[1]);
  }

  // Determine data type of output image.  Doing this "right" would
  // potentially take a large amount of memory, so just do a best estimate.
  // Assuming the Dal types will be adjusted on a machine type basis, it
  // should be safe to test on hard coded maximum values.

  Dal::DataType imageType = Dal::Int8;

  if( booleanParameter("withimagedatatype") ){

    std::string imagedatatype = stringParameter("imagedatatype");
    if( imagedatatype == "Int8" ) imageType = Dal::Int8;
    if( imagedatatype == "Int16" ) imageType = Dal::Int16;
    if( imagedatatype == "Int32" ) imageType = Dal::Int32;
    if( imagedatatype == "Real32" ) imageType = Dal::Real32;
    if( imagedatatype == "Real64" ) imageType = Dal::Real64;

    if(zcol){
      zDatCol->parent()->deleteColumn(zDatCol->name());
      zDatCol=reType(zCol, imageType);
    }

  }else{

    if(zcol){
      ColumnData *cZDat = zDatCol->data();
      real64 max=0, *zDat = cZDat->real64Data();
      unsigned long mRows = _intable->rows();
      for(unsigned long i=0; i<mRows; i++) max += fabs(zDat[i]);
      delete cZDat;

      Dal::DataType zColType = zCol->dataType();
      if(zIntCol){
	if( max < 127 && zColType == Dal::Int8 )
	  imageType = Dal::Int8;
	else if( max < 32767 && (zColType==Dal::Int8 || zColType==Dal::Int16) )
	  imageType = Dal::Int16;
	else 
	  imageType = Dal::Int32;

	zDatCol->parent()->deleteColumn(zDatCol->name());
	zDatCol=reType(zCol, imageType);
      }else{

	if( max < 3.402823466E+38 && zColType==Dal::Real32 ){
	  imageType = Dal::Real32;
	  zDatCol->parent()->deleteColumn(zDatCol->name());
	  zDatCol=reType(zCol, Dal::Real32);
	}else 
	  imageType = Dal::Real64;
      }

    }else{
      unsigned long nmax=_intable->rows();
      if( nmax < 127 )
	imageType = Dal::Int8;
      else if( nmax < 32767 )
	imageType = Dal::Int16;
      else 
	imageType = Dal::Int32;
    }
  }

  // Now initialize the output image ans set keyowrds

  Array * outImage = _dataSet->addArray("PRIMARY", imageType, size);

  _copyGlobalAttribs(outImage);

  // Write a few obvious keywords

  * outImage->addAttribute( "HDUCLASS",
			    "Format conforms to OGIP/GSFC conventions")
      = "OGIP";
  * outImage->addAttribute( "HDUCLAS1",
			    "File contains an Image")
      = "IMAGE";

  * outImage->addAttribute( "HDUCLAS2",
			    "Gross Image")
      = "TOTAL";

  * outImage->addAttribute( "HDUVERS1",
			    "Version of format")
      = "1.1.0";


  // Check for and set the WCS keywords

  std::vector<Column *> coord;
  coord.push_back(xCol);
  coord.push_back(yCol);

  std::vector<real64> coordMin;
  coordMin.push_back(xLimits[0]);
  coordMin.push_back(yLimits[0]);

  std::vector<real64> wcsBin;
  wcsBin.push_back(binSize[1]);
  wcsBin.push_back(binSize[0]);

  for(unsigned ico=0; ico<2; ico++){
    std::strstream typeName, pixName, valName, deltName, unitName, ltvName, ltmName;
    typeName << "CTYPE"<<ico+1<<ends;
    pixName << "CRPIX"<<ico+1<<ends;
    valName << "CRVAL"<<ico+1<<ends;    
    deltName << "CDELT"<<ico+1<<ends;
    unitName << "CUNIT"<<ico+1<<ends;
    ltvName << "LTV"<<ico+1<<ends;
    ltmName << "LTM"<<ico+1<<"_"<<ico+1<<ends;

    if( ! coord[ico]->hasAttribute("TCTYP") ||
	! coord[ico]->hasAttribute("TCRPX") ||
	! coord[ico]->hasAttribute("TCRVL") ||
	! coord[ico]->hasAttribute("TCDLT") ){
      errstr << "No WCS information available for image column "<<coord[ico]->name()<<"." <<Sas::warning("NoWCS");

      // Add WCS info based on the columns the image was extracted from.
      
      * outImage->addAttribute(typeName.str(), "WCS coordinate name") = coord[ico]->name();
      * outImage->addAttribute(pixName.str(), "WCS reference pixel") = 1;
      * outImage->addAttribute(valName.str(), "WCS reference pixel value") = coordMin[ico] + wcsBin[ico]*0.5;
      * outImage->addAttribute(deltName.str(), "WCS pixel size") = wcsBin[ico];

      * outImage->addAttribute(ltvName.str()) = coordMin[ico]-wcsBin[ico]*0.5;
      * outImage->addAttribute(ltmName.str()) = wcsBin[ico];
      continue;
    }

    outImage->addAttribute(coord[ico]->attribute("TCTYP"))->rename(typeName.str());

    real64 pixVal = coord[ico]->attribute("TCRPX")->asReal();
    * outImage->addAttribute(pixName.str(), "WCS reference pixel") = (pixVal-coordMin[ico])/wcsBin[ico] + 0.5;

    real64 value = coord[ico]->attribute("TCRVL")->asReal();
    outImage->addAttribute(coord[ico]->attribute("TCRVL"))->rename(valName.str());

    if(coord[ico]->hasAttribute("TCUNI")) outImage->addAttribute(coord[ico]->attribute("TCUNI"))->rename(unitName.str());

    real64 deltVal = coord[ico]->attribute("TCDLT")->asReal();
    * outImage->addAttribute(deltName.str(), "WCS pixel size") = deltVal*wcsBin[ico];

    //    * outImage->addAttribute(ltvName.str()) = value-deltVal*(pixVal-coordMin[ico])+deltVal*wcsBin[ico]*0.5;
    //    * outImage->addAttribute(ltmName.str()) = deltVal*wcsBin[ico];
    * outImage->addAttribute(ltvName.str()) = (pixVal-coordMin[ico])/wcsBin[ico] + 0.5 - value/(deltVal*wcsBin[ico]);
    * outImage->addAttribute(ltmName.str()) = 1/(deltVal*wcsBin[ico]);
  }
  * outImage->addAttribute("LTM1_2") = 0;
  * outImage->addAttribute("LTM2_1") = 0;

  // Accumulate the image.

  ArrayData * cImageData = outImage->data();

  // Set up type converted columns

  Column *flagDatCol;

  ColumnData *cFlagData = 0, *cZData = 0, *cXData=xDatCol->data(), *cYData=yDatCol->data();

  int32 *flagData = 0;
  real64 *xData=cXData->real64Data(), *yData=cYData->real64Data();

  if(! _destruct){
    flagDatCol = reType(_intable->column(_flagCol), Dal::Int32);
    cFlagData = flagDatCol->data();
    flagData = cFlagData->int32Data();
  }

  int ix, iy;

  if(zcol) cZData = zDatCol->data();

#define CODE(mtype)                                                      \
  {                                                                      \
    mtype *zData = 0, *imageData = cImageData->mtype##Data();            \
    for(unsigned im=0; im<size[0]*size[1]; im++) imageData[im]=0;        \
    if(zcol) zData = cZData->mtype##Data();                              \
									 \
    unsigned long mRows = _intable->rows();                              \
    for(unsigned long irow=0; irow < mRows; irow++){                     \
      if( ! _destruct && ((flagData[irow] & _flagBit) == 0)) continue;   \
									 \
      if( xData[irow]<xLimits[0] || xData[irow]>xLimits[1] )continue;    \
      if( yData[irow]<yLimits[0] || yData[irow]>yLimits[1] )continue;    \
									 \
      ix = (int)floor( (xData[irow]-xLimits[0])/binSize[1] );            \
      iy = (int)floor( (yData[irow]-yLimits[0])/binSize[0] );            \
									 \
      if(zcol){                                                          \
	imageData[iy*size[1]+ix] += zData[irow];                         \
      }else{                                                             \
	imageData[iy*size[1]+ix] += 1;                                   \
      }                                                                  \
    }                                                                    \
  break;                                                                 \
  }

  switch(imageType){

  case Dal::Int32  : CODE(int32)
		      
  case Dal::Int16  : CODE(int16)

  case Dal::Int8   : CODE(int8)

  case Dal::Real32 : CODE(real32)

  case Dal::Real64 : CODE(real64)

  default :
    errstr << "Image type is not of a supported data type." << Sas::error("ImageType");
  }

  #undef CODE

  delete cImageData;

  delete cXData;
  delete cYData;

  if(zcol) delete cZData;

  if(! _destruct) delete cFlagData;

  // Write data subspace info
  if (writeDss) {
     DataSubspace dss(_intable);
     outImage << dss;
  }

  // Calculate exposure keyword

  _getExposure(outImage);

  // Done
}


void Image::_getExposure(Block *productBlock){

  if( !_intable->parent()->hasAttribute("TELESCOP") ||
      _intable->parent()->attribute("TELESCOP")->asString() != "XMM" ){
    errstr << "Not XMM data.  Cannot calculate EXPOSURE keyword for output products."
	   << Sas::warning("NotXmmProduct");
    return;
  }
  
  std::string instrument = _intable->parent()->attribute("INSTRUME")->asString();
  
  if(instrument=="RGS1" || instrument=="RGS2"){

    if( !_intable->hasAttribute("EXPOSURE") ){
      errstr << "No EXPOSURE keyword in RGS event list.  " 
	     <<	"Cannot write exposure value to the product file."
	     << Sas::warning("NoRGSExposure");
      return;
    }

    productBlock->addAttribute(_intable->attribute("EXPOSURE"));

  }else if(instrument=="EMOS1" || instrument=="EMOS2" || instrument=="EPN"){

    real64 exposure=0;

    for(unsigned iccd=1; iccd<=17; iccd++){

      std::strstream keyName;
      keyName << "ONTIME";
      keyName.width(2);
      keyName.fill('0');
      keyName << iccd << ends;

      if( _intable->hasAttribute(keyName.str()) ){
	real64 expval = productBlock->addAttribute( _intable->attribute(keyName.str()) )->asReal();
	if( expval>exposure ) exposure = expval;
      }
    }

    *productBlock->addAttribute("EXPOSURE", 
				 "max of ONTIMEnn values") = exposure;
  }

  return;
}

vector<double> Image::_radec2xy(vector<double> worldVec, Column *Xcol, Column *Ycol){

  double phi, theta, imgcrd[2], pixcrd[2];

  char ctype[2][9];
  strcpy(ctype[0], Xcol->attribute("TCTYP")->asString().c_str());
  strcpy(ctype[1], Ycol->attribute("TCTYP")->asString().c_str());

  struct wcsprm* wcs = new struct wcsprm;
  wcs->flag = 0;

  double world[2];
  world[0] = worldVec[0];
  world[1] = worldVec[1];
  
  double crval[2];
  crval[0] = Xcol->attribute("TCRVL")->asReal();
  crval[1] = Ycol->attribute("TCRVL")->asReal();

  struct celprm* cel = new struct celprm;
  cel->flag = 0;
  cel->ref[0] = Xcol->attribute("TCRVL")->asReal();
  cel->ref[1] = Ycol->attribute("TCRVL")->asReal();
  cel->ref[2] = cel->ref[3] = 999;

  struct prjprm* prj = new struct prjprm;
  prj->flag = 0;
  prj->r0 = 0;
  for(unsigned i=0; i<10; i++) prj->p[i] = 0;

  double crpix[2], m[4], cdelt[2];
  crpix[0] = Xcol->attribute("TCRPX")->asReal();
  crpix[1] = Ycol->attribute("TCRPX")->asReal();
  m[0] = m[3] = 1;
  m[1] = m[2] = 0;
  cdelt[0] = Xcol->attribute("TCDLT")->asReal();
  cdelt[1] = Ycol->attribute("TCDLT")->asReal();

  struct linprm* lin = new struct linprm;
  lin->flag = 0;
  lin->naxis = 2;
  lin->crpix = crpix;
  lin->pc = m;
  lin->cdelt = cdelt;

  int result = wcsfwd(ctype, wcs, world, crval, cel, &phi, &theta, prj, imgcrd, lin, pixcrd);
  if(result != 0)
    errstr << "Cannot transform wcs coordinates (" << worldVec[0] << "," << worldVec[1]
	   << ") to pixel coordinates.  wcsfwd error " << result << Sas::error("WcsTransform");

  vector<double> pixVec;
  pixVec.push_back(pixcrd[0]);
  pixVec.push_back(pixcrd[1]);

  delete wcs;
  delete cel;
  free(lin->piximg);
  free(lin->imgpix);
  delete lin;
  return pixVec;
}
